/*
 * MyTextBox - a custom control that implements the core functionality
 * of a standard TextBox with support for proportional space fonts.
 * 
 * Ben Bederson
 * April 2005
 * (c) University of Maryland
 */

using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Text;
using System.Data;
using System.Windows.Forms;
using System.Threading;

namespace MyTextBox {
	public class MyTextBox : System.Windows.Forms.UserControl {
		#region Fields
		private const string DISPLAYABLE_CHARS = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMOPQRSTUVWXYZ1234567890`~!@#$%^&*()_+-=[]\\{}|;':\",./<>? ";
		private System.ComponentModel.Container components = null;
		private ArrayList lines;
		private Point cursorPos;
		private Font font;
		private bool cursorVisible = true;
		private Thread cursorThread;
		#endregion

		#region Constructor
		public MyTextBox() {
			InitializeComponent();

			lines = new ArrayList();
			lines.Add("");
			font = new Font("Times", 12);
			cursorPos = new Point(0, 0);
			KeyPress += new KeyPressEventHandler(MyTextBox_KeyPress);
			Layout += new LayoutEventHandler(MyTextBox_Layout);
			Load += new EventHandler(MyTextBox_Load);
			HandleDestroyed += new EventHandler(MyTextBox_HandleDestroyed);

			SetStyle(ControlStyles.DoubleBuffer, true);
			SetStyle(ControlStyles.UserPaint, true);
			SetStyle(ControlStyles.AllPaintingInWmPaint, true);
		}
		#endregion

		#region Properties
		public override Font Font {
			get { return font; }
			set {
				font = value;
				LayoutMyTextBox();
			}
		}
		#endregion

		#region Rendering
		// Render the text box
		protected override void OnPaint(PaintEventArgs e) {
			e.Graphics.Clear(Color.White);
			StringFormat format = new StringFormat(StringFormatFlags.MeasureTrailingSpaces);

			// Draw text
			int y = 0;
			foreach (string line in lines) {
				string lineToRender = line.Replace("\n", "");  // Don't render '\n's
				e.Graphics.DrawString(lineToRender, font, Brushes.Black, 0, y);
				int dy = (int)e.Graphics.MeasureString(lineToRender, font, new PointF(0, 0), format).Height;
				if (dy == 0) {
					dy = (int)e.Graphics.MeasureString(" ", font, new PointF(0, 0), format).Height;
				}
				y += dy;
			}

			// Draw cursor (after calculating where it goes)
			if (cursorVisible) {
				string cursorLine = (lines[cursorPos.Y] as string).Replace("\n", "");
				string cursorSubLine = cursorLine.Substring(0, cursorPos.X);
				SizeF subLineSize = new SizeF(0, 0);
				if (cursorSubLine.Length > 0) {
					CharacterRange[] ranges = new CharacterRange[] {new CharacterRange(0, cursorSubLine.Length)};
					format.SetMeasurableCharacterRanges(ranges);
					RectangleF layoutRect = new RectangleF(0, 0, 10000, 100);
					Region[] regions = e.Graphics.MeasureCharacterRanges(cursorSubLine, font, layoutRect, format);
					RectangleF regionRect = regions[0].GetBounds(e.Graphics);
					subLineSize = new SizeF(regionRect.Width, regionRect.Height);
				}
				if (subLineSize.Height == 0) {
					subLineSize.Height = e.Graphics.MeasureString(" ", font, new Point(0, 0), format).Height; // Take care of special case of empty line
				}
				int cursorX = (int)subLineSize.Width;
				int cursorY = (int)(cursorPos.Y * subLineSize.Height);
				e.Graphics.DrawLine(Pens.Black, cursorX, cursorY, cursorX, cursorY + subLineSize.Height);
			}
		}
		#endregion

		#region Key handling
		// Handle displayable keys
		private void MyTextBox_KeyPress(object sender, KeyPressEventArgs e) {
			char c = e.KeyChar;
			if (DISPLAYABLE_CHARS.IndexOf(c) != -1) {
				string line = (string)lines[cursorPos.Y];
				line = line.Insert(cursorPos.X, c.ToString());
				lines[cursorPos.Y] = line;
				cursorPos.X++;
			}

			LayoutMyTextBox();
		}

		// Handle special keys
		protected override bool ProcessCmdKey(ref Message msg, Keys keyData) {
			string line = (string)lines[cursorPos.Y];

			switch (keyData) {
				case Keys.Left:
					cursorPos.X--;
					if (cursorPos.X < 0) {  // If past beginning of line
						cursorPos.Y--;
						if (cursorPos.Y < 0) { // If past top of text
							cursorPos.X = 0;
							cursorPos.Y = 0;
						} else {
							cursorPos.X = (lines[cursorPos.Y] as string).Length;
						}
					}
					break;
				case Keys.Right:
					cursorPos.X++;
					if ((cursorPos.X > line.Length) ||
						((line[line.Length-1] == '\n') && (cursorPos.X == line.Length))) { // If past end of line
						if (cursorPos.Y >= (lines.Count - 1)) { // If past bottom of text
							cursorPos.X--;
						} else {
							cursorPos.X = 0;
							cursorPos.Y++;
						}
					}
					break;
				case Keys.Up:
					if (cursorPos.Y > 0) { // If not at top of text
						// First calculate current line width
						string subLine = line.Substring(0, cursorPos.X);
						Graphics g = CreateGraphics();
						StringFormat format = new StringFormat(StringFormatFlags.MeasureTrailingSpaces);
						int width = (int)g.MeasureString(subLine, font, new Point(0, 0), format).Width;

						// Now on previous line, calculate best position for cursor
						cursorPos.Y--;
						line = (string)lines[cursorPos.Y];
						cursorPos.X = 0;
						int currentWidth = 0;
						foreach (char c in line) {
							int charWidth = (int)g.MeasureString(c.ToString(), font, new Point(0, 0), format).Width;
							if (currentWidth + (charWidth / 2) < width) {
								cursorPos.X++;
								currentWidth += charWidth;
							} else {
								break;
							}
						}
					}
					break;
				case Keys.Down:
					if (cursorPos.Y < (lines.Count - 1)) { // If not at bottom of text
						// First calculate current line width
						string subLine = line.Substring(0, cursorPos.X);
						Graphics g = CreateGraphics();
						StringFormat format = new StringFormat(StringFormatFlags.MeasureTrailingSpaces);
						int width = (int)g.MeasureString(subLine, font, new Point(0, 0), format).Width;

						// Now on next line, calculate best position for cursor
						cursorPos.Y++;
						line = (string)lines[cursorPos.Y];
						cursorPos.X = 0;
						int currentWidth = 0;
						foreach (char c in line) {
							int charWidth = (int)g.MeasureString(c.ToString(), font, new Point(0, 0), format).Width;
							if (currentWidth + (charWidth / 2) < width) {
								cursorPos.X++;
								currentWidth += charWidth;
							} else {
								break;
							}
						}
					}
					break;
				case Keys.Delete:
					if (cursorPos.X < line.Length) { // If not at end of line
						lines[cursorPos.Y] = line.Substring(0, cursorPos.X) + line.Substring(cursorPos.X + 1);
					} else { // Join w/ next line here
						if (cursorPos.Y < (lines.Count - 1)) { // If not at bottom line
							string newLine = line + (lines[cursorPos.Y + 1] as string);
							lines[cursorPos.Y] = newLine;
							lines.RemoveAt(cursorPos.Y + 1);
						}
					}
					break;
				case Keys.Back:
					if (cursorPos.X > 0) {
						lines[cursorPos.Y] = line.Substring(0, cursorPos.X-1) + line.Substring(cursorPos.X);
						cursorPos.X--;
					} else { // Join w/ previous line here
						if (cursorPos.Y > 0) { // If not at top line
							string newLine = (lines[cursorPos.Y - 1] as string) + line;
							cursorPos.X = newLine.Length;
							cursorPos.Y--;
							lines[cursorPos.Y] = newLine;
							lines.RemoveAt(cursorPos.Y + 1);
						}
					}
					break;
				case Keys.Home:
					cursorPos.X = 0;
					break;
				case Keys.End:
					cursorPos.X = line.Length;
					break;
				case Keys.Enter:
					// First add a '\n' to the current line after the cursor
					line = line.Insert(cursorPos.X, "\n");
					lines[cursorPos.Y] = line;
					cursorPos.X++;

					// Then insert a new blank line if was at the end of a line
					if (cursorPos.X == line.Length) {
						cursorPos.X = 0;
						cursorPos.Y++;
						lines.Insert(cursorPos.Y, "");
					} else {
						// Or split the current line if was in the middle of a line
						string line1 = line.Substring(0, cursorPos.X);
						string line2 = line.Substring(cursorPos.X);
						lines[cursorPos.Y] = line1;
						lines.Insert(cursorPos.Y+1, line2);
						cursorPos.X = 0;
						cursorPos.Y++;
					}
					break;
			}

			// Don't allow cursor to go after a '\n'
			if (cursorPos.X > 0) {
				line = (string)lines[cursorPos.Y];
				if (line[cursorPos.X-1] == '\n') {
					cursorPos.X--;
				}
			}
			
			LayoutMyTextBox();

			return base.ProcessCmdKey (ref msg, keyData);
		}
		#endregion

		#region Cursor blink management
		// Create cursor blink thread when textbox loads
		private void MyTextBox_Load(object sender, EventArgs e) {
			cursorThread = new Thread(new ThreadStart(BlinkCursor));
			cursorThread.Start();
		}

		// Stop cursor blink thread when textbox is destroyed
		private void MyTextBox_HandleDestroyed(object sender, EventArgs e) {
			cursorThread.Abort();
			cursorThread = null;
		}

		// This method runs in a special thread to blink the cursor
		protected void BlinkCursor() {
			do {
				cursorVisible = !cursorVisible;
				Invalidate();
				Thread.Sleep(500);
			} while (true);
		}
		#endregion

		#region Layout management
		private void MyTextBox_Layout(object sender, LayoutEventArgs e) {
			LayoutMyTextBox();
		}

		protected void LayoutMyTextBox() {
			ArrayList newLines = new ArrayList();

			Graphics g = CreateGraphics();
			StringFormat format = new StringFormat(StringFormatFlags.MeasureTrailingSpaces);
			int width = ClientRectangle.Width;

			// Perform word wrap calculation by creating new wrapped list of lines
			// KNOWN BUG: Doesn't handle multiple sequential spaces
			int currentWidth = 0;
			string currentLine = "";
			foreach (string line in lines) {
				string[] words = line.Split(new char[] {' '});
				foreach (string word in words) {
					int wordWidth = (int)g.MeasureString(word, font, new Point(0, 0), format).Width;
					if ((currentWidth + wordWidth) < width) {
						if (currentLine.Length == 0) {
							currentLine = word;
						} else {
							currentLine = currentLine + " " + word;
						}
						currentWidth = (int)g.MeasureString(currentLine, font, new Point(0, 0), format).Width;
					} else {
						newLines.Add(currentLine);
						currentLine = word;
						currentWidth = 0;
					}
					// Handle '\n'
					if (word.EndsWith("\n")) {
						newLines.Add(currentLine);
						currentLine = "";
						currentWidth = 0;
					}
				}
			}
			newLines.Add(currentLine);
			lines = newLines;

			// KNOWN BUG: Avoid cursor position errors - but it should calculate its new position
			// so that it was at the same point in the text.
			if (cursorPos.Y >= lines.Count) {
				cursorPos.Y = lines.Count - 1;
			}
			string currLine = (string)lines[cursorPos.Y];
			if (cursorPos.X > currLine.Length) {
				cursorPos.X = currLine.Length;
				if ((currLine.Length > 0) && (currLine[currLine.Length-1] == '\n') && (cursorPos.X == currLine.Length)) { // If after '\n'
					cursorPos.X--;
				}
			}

			Invalidate();
		}
		#endregion

		#region Dispose
		protected override void Dispose( bool disposing ) {
			if( disposing ) {
				if(components != null) {
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}
		#endregion

		#region Component Designer generated code
		/// <summary> 
		/// Required method for Designer support - do not modify 
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent() {
			components = new System.ComponentModel.Container();
		}
		#endregion
	}
}